/*
 * RenderPipeline.java
 */

package de.ba_ravensburg.gdv.controller;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import de.ba_ravensburg.gdv.model.Scene;
import de.ba_ravensburg.gdv.model.Triangle;
import de.ba_ravensburg.gdv.view.View3DPanel;

/**
 * The <i>RenderPipeline</i> is a pipeline inside of which the triangles of a
 * scene are transformed into a form suitable for displaying them on the screen.
 * 
 * @author Ansgar Wiechers
 */
public class RenderPipeline {

  private final int DIMENSION = Scene.DIMENSION;
  private final int X = Scene.X;
  private final int Y = Scene.Y;
  private final int Z = Scene.Z;
  private final int HORIZONTAL = Triangle.HORIZONTAL;
  private final int VERTICAL = Triangle.VERTICAL;
  
  private LinkedList<Triangle> queue;
  private LinkedList<Triangle2D> triangles2D;
  private double aperture;
  private double alpha;
  private double size = View3DPanel.SIZE / 2.0;

  /**
   * Constructs a new empty rendering pipeline.
   */
  RenderPipeline() {
    queue = new LinkedList<Triangle>();
    triangles2D = new LinkedList<Triangle2D>();
  }

  /**
   * Put the given triangles into the rendering pipeline. Each triangle that is
   * completely outside of the camera's view port or has its frontside turned
   * away from the camera is not put into the pipeline.
   * <p>
   * <table><tr>
   * <td>Pre:  <td><i>triangles</i> != null<br>
   *               0 &lt; <i>theAperture</i> &lt;= 180
   * </tr><tr>
   * <td>Post: <td>Each triangle that is not definitely outside the view port
   *               (i.e. behind the camera or right/left/above/below the view
   *               port) and is facing towards the camera is put into the queue
   *               for rendering.
   * </tr></table>
   * 
   * @param triangles a set of triangles to be put into the rendering pipeline.
   */
  void put(HashSet triangles, double theAperture)
    throws IllegalArgumentException {
    Iterator it;
    Triangle theTriangle;
    double phiA, phiB, phiC, rhoA, rhoB, rhoC;
    
    if (theAperture <= 0.0 || theAperture > 180.0 || triangles == null)
      throw new IllegalArgumentException();
    
    aperture = theAperture;
    alpha = Math.tan(theAperture / 180.0 * Math.PI);
    queue.clear();
    
    it = triangles.iterator();
    while (it.hasNext()) {
      theTriangle = (Triangle)it.next();
      
      // Clipping
      if (isBehindCamera(theTriangle))
        continue;
      
      phiA = theTriangle.getAngleOfA(HORIZONTAL);
      phiB = theTriangle.getAngleOfB(HORIZONTAL);
      phiC = theTriangle.getAngleOfC(HORIZONTAL);
      
      if (isLeftOfCamera(phiA, phiB, phiC))
        continue;
      
      if (isRightOfCamera(phiA, phiB, phiC))
        continue;
      
      rhoA = theTriangle.getAngleOfA(VERTICAL);
      rhoB = theTriangle.getAngleOfB(VERTICAL);
      rhoC = theTriangle.getAngleOfC(VERTICAL);
      
      if (isAboveCamera(rhoA, rhoB, rhoC))
        continue;
      
      if (isBelowCamera(rhoA, rhoB, rhoC))
        continue;
      
      // Backface culling
      if (isTurnedAway(theTriangle))
        continue;
      
      put(theTriangle);
    }
  }
  
  /**
   * Get a sorted list of 2D triangles from the rendering pipeline. The
   * triangles in this list are sorted descending by their distance from the
   * camera.
   * 
   * @return list of 2D triangles to be painted on the screen.
   */
  LinkedList get() {
    Triangle triangle;
    double[][] points = new double[3][3];
    double[] point;
    int[] x;
    int[] y;
    int npoints, i, j;
    
    double k, x0, z0, x1, z1;
    double dy = 0.01;
    
    triangles2D.clear();
    
    while (!queue.isEmpty()) {
      triangle = (Triangle)queue.removeLast();
      
      points[0] = triangle.getPointA();
      points[1] = triangle.getPointB();
      points[2] = triangle.getPointC();
      
      if (points[0][Y] > 0.0 && points[1][Y] > 0.0 && points[2][Y] > 0.0) {
        npoints = 3;
        x = new int[npoints];
        y = new int[npoints];
        
        for (i = 0; i < npoints; ++i) {
          x[i] = coordinate2D(points[i][X], points[i][Y]);
          y[i] = coordinate2D(-points[i][Z], points[i][Y]);
        }
      } else {
        /* Sort the points ascending by their y-coordinate. If the y-coordinate
         * of the middle point is not greater than zero, we have two points with
         * y <= 0.0, otherwise just one.
         * Because we have at most three exchanges, Bubblesort is sufficient.
         */
        for (j = 0; j < 2; ++j)
          for (i = 0; i < 2 - j; ++i)
            if (points[i][Y] > points[i+1][Y]) {
              point = points[i];
              points[i] = points[i+1];
              points[i+1] = point;
            }
        
        if (points[1][Y] <= 0.0) {
          // Only one point has a y-coordinate greater than zero.
          npoints = 3;
          x = new int[npoints];
          y = new int[npoints];
          
          /* Calculate the intersection point of the straight line through
           * point[0] and point[2] with a plane which is parallel to the
           * xz-plane with a distance of dy.
           */
          k = (dy - points[0][Y]) / (points[2][Y] - points[0][Y]);
          x0 = points[0][X] + k * (points[2][X] - points[0][X]);
          z0 = points[0][Z] + k * (points[2][Z] - points[0][Z]);
          
          /* Calculate the intersection point of the straight line through
           * point[1] and point[2] with a plane which is parallel to the
           * xz-plane with a distance of dy.
           */
          k = (dy - points[1][Y]) / (points[2][Y] - points[1][Y]);
          x1 = points[1][X] + k * (points[2][X] - points[1][X]);
          z1 = points[1][Z] + k * (points[2][Z] - points[1][Z]);
        } else {
          // Two points have a y-coordinate greater than zero.
          npoints = 4; // Yes, this triangle has four angles. So what?
          x = new int[npoints];
          y = new int[npoints];
          
          /* Calculate the intersection point of the straight line through
           * point[0] and point[2] with a plane which is parallel to the
           * xz-plane with a distance of dy.
           */
          k = (dy - points[0][Y]) / (points[2][Y] - points[0][Y]);
          x0 = points[0][X] + k * (points[2][X] - points[0][X]);
          z0 = points[0][Z] + k * (points[2][Z] - points[0][Z]);
          
          /* Calculate the intersection point of the straight line through
           * point[0] and point[1] with a plane which is parallel to the
           * xz-plane with a distance of dy.
           */
          k = (dy - points[0][Y]) / (points[1][Y] - points[0][Y]);
          x1 = points[0][X] + k * (points[1][X] - points[0][X]);
          z1 = points[0][Z] + k * (points[1][Z] - points[0][Z]);
          
          x[2] = coordinate2D(points[1][X], points[1][Y]);
          y[2] = coordinate2D(-points[1][Z], points[1][Y]);
        }
        
        x[0] = coordinate2D(x0, dy);
        y[0] = coordinate2D(-z0, dy);
        
        x[1] = coordinate2D(x1, dy);
        y[1] = coordinate2D(-z1, dy);
        
        x[npoints-1] = coordinate2D(points[2][X], points[2][Y]);
        y[npoints-1] = coordinate2D(-points[2][Z], points[2][Y]);
      }
      
      triangles2D.addFirst(new Triangle2D(x, y, npoints, triangle.getColor()));
    }
    
    return triangles2D;
  }
  
  /**
   * Put given triangle <i>theTriangle</i> into the queue. The queue is sorted
   * descending by distance from the camera position and <i>theTriangle</i> is
   * inserted after the last triangle inside the queue that has a greater
   * distance from the camera than <i>theTriangle</i>.
   * 
   * @param theTriangle a triangle.
   */
  private void put(Triangle theTriangle) {
    int pos, size;
    double distance;

    size = queue.size();
    distance = theTriangle.getDistance();
    
    // P:    All triangles before position pos in the queue have greater
    //       distance than theTriangle.
    // Stop: end of queue is reached or
    //       distance(triangle[pos]) <= distance(theTriangle)
    for (pos = 0;
         pos < size && ((Triangle)queue.get(pos)).getDistance() > distance;
         ++pos);
    
    queue.add(pos, theTriangle);
  }
  
  /**
   * Determine, whether the given triangle is behind the camera or not.
   * 
   * @param theTriangle a triangle.
   * @return true if each point of the triangle is behind the camera, else
   *         false.
   */
  private boolean isBehindCamera(Triangle theTriangle) {
    return (theTriangle.getPointA()[Y] < 0.0 &&
            theTriangle.getPointB()[Y] < 0.0 &&
            theTriangle.getPointC()[Y] < 0.0) ? true : false;
  }
  
  /**
   * Determine whether a triangle is completely left outside the view port.
   * 
   * @param phiA polar coordinate angle of point A (horizontal).
   * @param phiB polar coordinate angle of point B (horizontal).
   * @param phiC polar coordinate angle of point C (horizontal).
   * @return true if each point of the triangle is to the left of the left
   *         perimeter of the view port, else false.
   */
  private boolean isLeftOfCamera(double phiA, double phiB, double phiC) {
    if (phiA < 180.0 - aperture && phiA > -aperture)
      return false;
    
    if (phiB < 180.0 - aperture && phiB > -aperture)
      return false;
    
    if (phiC < 180.0 - aperture && phiC > -aperture)
      return false;
    
    return true;
  }
  
  /**
   * Determine whether a triangle is completely right outside the view port.
   * 
   * @param phiA polar coordinate angle of point A (horizontal).
   * @param phiB polar coordinate angle of point B (horizontal).
   * @param phiC polar coordinate angle of point C (horizontal).
   * @return true if each point of the triangle is to the right of the right
   *         perimeter of the view port, else false.
   */
  private boolean isRightOfCamera(double phiA, double phiB, double phiC) {
    if (phiA > aperture - 180.0 && phiA < aperture)
      return false;
    
    if (phiB > aperture - 180.0 && phiB < aperture)
      return false;
    
    if (phiC > aperture - 180.0 && phiC < aperture)
      return false;
    
    return true;
  }
  
  /**
   * Determine whether a triangle is completely above outside the view port.
   * 
   * @param rhoA polar coordinate angle of point A (vertical).
   * @param rhoB polar coordinate angle of point B (vertical).
   * @param rhoC polar coordinate angle of point C (vertical).
   * @return true if each point of the triangle is above the upper perimeter of
   *         the view port, else false.
   */
  private boolean isAboveCamera(double rhoA, double rhoB, double rhoC) {
    if (rhoA> aperture - 180.0 && rhoA < aperture)
      return false;
    
    if (rhoB > aperture - 180.0 && rhoB < aperture)
      return false;
    
    if (rhoC > aperture - 180.0 && rhoC < aperture)
      return false;
    
    return true;
  }
  
  /**
   * Determine whether a triangle is completely below outside the view port.
   * 
   * @param rhoA polar coordinate angle of point A (vertical).
   * @param rhoB polar coordinate angle of point B (vertical).
   * @param rhoC polar coordinate angle of point C (vertical).
   * @return true if each point of the triangle is below the lower perimeter of
   *         the view port, else false.
   */
  private boolean isBelowCamera(double rhoA, double rhoB, double rhoC) {
    if (rhoA < 180.0 - aperture && rhoA > -aperture)
      return false;
    
    if (rhoB < 180.0 - aperture && rhoB > -aperture)
      return false;
    
    if (rhoC < 180.0 - aperture && rhoC > -aperture)
      return false;
    
    return true;
  }
  
  /**
   * Determine whether a triangle is facing towards the camera or not (only one
   * "side" of the triangle is visible!). This is used for backface culling.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>true
   * </tr><tr>
   * <td>Post: <td>If the visible side of the triangle is turned towards the
   *               camera, true is returned, else false is returned.
   * </tr></table>
   * 
   * @param theTriangle a triangle.
   * @return true if the angle between the triangle's point A and its normal
   *         vector is less or equal 90, else false.
   */
  private boolean isTurnedAway(Triangle theTriangle) {
    double[] a, b, c;
    double[] n = new double[DIMENSION];
    double p;
    
    a = theTriangle.getPointA();
    b = theTriangle.getPointB();
    c = theTriangle.getPointC();
    
    // Calculate the normal vector of the triangle.
    n[X] = (b[Y] - a[Y]) * (c[Z] - a[Z]) - (b[Z] - a[Z]) * (c[Y] - a[Y]);
    n[Y] = (b[Z] - a[Z]) * (c[X] - a[X]) - (b[X] - a[X]) * (c[Z] - a[Z]);
    n[Z] = (b[X] - a[X]) * (c[Y] - a[Y]) - (b[Y] - a[Y]) * (c[X] - a[X]);
    
    /* Whether a triangle is turned away from the camera or not is determined
     * by the direction that the projection of the triangle's normal vector
     * on the position vector of point A has. If the direction is greater than
     * zero, then the triangle is facing towards the camera.
     * 
     *    a * n = |a| * |n| * cos(phi)
     * 
     *    a * n
     * => ----- = |n| * cos(phi) = p
     *     |a|
     * 
     * p is the length of the projection of n on a. Its direction is determined
     * by the unit vector of a:
     * 
     *              a
     *    n' = p * ---
     *             |a|
     * 
     *         a * n    a    a * n
     * => n' = ----- * --- = ----- * a = p' * a
     *          |a|    |a|    |a|
     * 
     * p' >= 0 => a and projection of n on a are parallel
     *         => triangle is facing towards the camera
     * p' < 0  => a and projection of n on a are antiparallel
     *         => triangle is *not* facing towards the camera
     * 
     * p' is easier to calculate than p, because we can avoid calculating the
     * square root for the absolute value of a, so we will use p' rather than p
     * to determine if the triangle is facing towards the camera.
     */
    p = (a[X] * n[X] + a[Y] * n[Y] + a[Z] * n[Z]) /
        (a[X] * a[X] + a[Y] * a[Y] + a[Z] * a[Z]);
    
    if (p <= 0.0)
      return true;
    
    return false;
  }
  
  /**
   * Calculate the corresponding 2D coordinate for the given 3D coordinate of a
   * point.
   * <p>
   * <table><tr>
   * <td>Pre:  <td>distance &gt; 0
   * </tr><tr>
   * <td>Post: <td>true
   * </tr></table>
   * 
   * @param value    a value.
   * @param distance the distance of the 3D point from the camera.
   * @return the 2D coordinate corresponding to 3D coordinate <i>value</i>.
   * @throws IllegalArgumentException if <i>distance</i> is less or equal zero.
   */
  private int coordinate2D(double value, double distance)
    throws IllegalArgumentException {
    double val;
    
    if (distance <= 0.0)
      throw new IllegalArgumentException();
    
    val = Math.round((1.0 + value / distance / alpha) * size);
    if (val > Integer.MAX_VALUE)
      return Integer.MAX_VALUE;
    else if (val < Integer.MIN_VALUE)
      return Integer.MIN_VALUE;
    else
      return (int)val;
  }

}
